#include "customfn.h"

#include <rbx/memory/functions/functions.h>
#include <rbx/memory/memdata/memdata.h>
#include <crypto/crypto.h>
#include <stdexcept>
#include <xorstr.h>

namespace temple::customfn
{
	using namespace memdata::lua::types;

	namespace crypt
	{
		struct buffer
		{
			unsigned long long size{};
			unsigned char* p;

			buffer() = delete;

			explicit buffer(std::size_t size) : p{ new unsigned char[size] }, size{ size }
			{}

			auto* data()
			{
				return p;
			}

			char* cdata()
			{
				return reinterpret_cast<char*>(p);
			}

			auto& length()
			{
				return size;
			}

			~buffer()
			{
				if (p)
					delete[] p;
			}
		};

		void decode_aeskey(buffer& bin_key, const char* key, std::size_t key_len)
		{
			switch (key_len)
			{
			case (crypto_aead_aes256gcm_KEYBYTES * 2):
			{
				auto decoded = crypto::from_hex(std::string(key, key_len));

				if (decoded.size() != bin_key.length())
					throw std::runtime_error{ xorstr("Invalid key format for AES-256 GCM") };

				std::memcpy(bin_key.data(), decoded.data(), bin_key.length());
				break;
			}
			case crypto_aead_aes256gcm_KEYBYTES:
				std::memcpy(bin_key.data(), key, crypto_aead_aes256gcm_KEYBYTES);
				break;
			default:
				throw std::runtime_error{ xorstr("Invalid key length for AES-256 GCM") };
			}
		}

		int encrypt(std::uintptr_t thread)
		{
			if (!crypto_aead_aes256gcm_is_available())
				throw std::runtime_error{ xorstr("AES-256 GCM is not supported on this processor") };

			auto data = get_argument(thread, 1, lua_tstring);
			auto pkey = get_argument(thread, 2, lua_tstring);

			const unsigned char* message = reinterpret_cast<const unsigned char*>(functions::lua::tostring(thread, 1));
			auto message_len = functions::lua::string::get_length(data->v.gcobj);

			const char* key = functions::lua::tostring(thread, 2);
			auto key_len = functions::lua::string::get_length(pkey->v.gcobj);

			buffer bin_key{ crypto_aead_aes256gcm_KEYBYTES };

			decode_aeskey(bin_key, key, key_len);

			buffer enc_buffer{ message_len + crypto_aead_aes256gcm_ABYTES + crypto_aead_aes256gcm_NPUBBYTES };

			auto* nonce = enc_buffer.data() + message_len + crypto_aead_aes256gcm_ABYTES;
			randombytes_buf(nonce, crypto_aead_aes256gcm_NPUBBYTES);

			unsigned long long l = message_len + crypto_aead_aes256gcm_ABYTES;
			crypto_aead_aes256gcm_encrypt(enc_buffer.data(), &l, message, message_len, nullptr, 0, nullptr, nonce, bin_key.data());

			functions::lua::pushlstring(thread, enc_buffer.cdata(), enc_buffer.length());

			return 1;
		}

		int decrypt(std::uintptr_t thread)
		{
			if (!crypto_aead_aes256gcm_is_available())
				throw std::runtime_error{ xorstr("AES-256 GCM is not supported on this processor") };

			auto data = get_argument(thread, 1, lua_tstring);
			auto pkey = get_argument(thread, 2, lua_tstring);

			const unsigned char* message = reinterpret_cast<const unsigned char*>(functions::lua::tostring(thread, 1));
			auto message_len = functions::lua::string::get_length(data->v.gcobj);

			const char* key = functions::lua::tostring(thread, 2);
			auto key_len = functions::lua::string::get_length(pkey->v.gcobj);

			if (message_len < crypto_aead_aes256gcm_ABYTES + crypto_aead_aes256gcm_NPUBBYTES)
				throw std::runtime_error{ xorstr("Invalid data size for AES-256 GCM") };

			buffer bin_key{ crypto_aead_aes256gcm_KEYBYTES };

			decode_aeskey(bin_key, key, key_len);

			auto* nonce = message + message_len - crypto_aead_aes256gcm_NPUBBYTES;

			buffer decrypted{ message_len - crypto_aead_aes256gcm_ABYTES - crypto_aead_aes256gcm_NPUBBYTES };

			if (crypto_aead_aes256gcm_decrypt(decrypted.data(), &decrypted.length(), 0, message, message_len - crypto_aead_aes256gcm_NPUBBYTES, nullptr, 0, nonce, bin_key.data()) != 0)
				throw std::runtime_error{ xorstr("AES-256 GCM Message forged!") };

			functions::lua::pushlstring(thread, decrypted.cdata(), decrypted.length());

			return 1;
		}

		int random(std::uintptr_t thread)
		{
			auto size_a = get_argument(thread, 1, lua_tnumber);
			auto size = functions::lua::tonumber(thread, 1);

			if (size < 0)
				throw std::runtime_error{ xorstr("size must not be lower than 0 or exceed 1024") };

			buffer rand_buf{ std::min<std::size_t>(functions::lua::tonumber(thread, 1), 1024) };

			randombytes_buf(rand_buf.data(), rand_buf.length());

			functions::lua::pushlstring(thread, rand_buf.cdata(), rand_buf.length());
			return 1;

		}

		namespace base64
		{
			int encode(std::uintptr_t thread)
			{
				auto data = get_argument(thread, 1, lua_tstring);
				auto data_ptr = reinterpret_cast<const unsigned char*>(functions::lua::tostring(thread, 1));
				auto data_len = functions::lua::string::get_length(data->v.gcobj);

				buffer encoded{ sodium_base64_ENCODED_LEN(data_len, sodium_base64_VARIANT_ORIGINAL) };

				sodium_bin2base64(encoded.cdata(), encoded.length(), data_ptr, data_len, sodium_base64_VARIANT_ORIGINAL);

				functions::lua::pushlstring(thread, encoded.cdata(), std::min<std::size_t>(encoded.length(), 1) - 1);
				return 1;
			}

			int decode(std::uintptr_t thread)
			{
				auto data = get_argument(thread, 1, lua_tstring);
				auto data_ptr = reinterpret_cast<const char*>(functions::lua::tostring(thread, 1));
				auto data_len = functions::lua::string::get_length(data->v.gcobj);

				buffer decoded{ data_len };
				std::size_t decoded_len{};

				if (sodium_base642bin(decoded.data(), data_len,
					data_ptr, data_len,
					nullptr, &decoded_len,
					nullptr, sodium_base64_VARIANT_ORIGINAL) != 0)
					throw std::runtime_error{ xorstr("Could not decode base64 string") };

				functions::lua::pushlstring(thread, decoded.cdata(), decoded_len);
				return 1;
			}
		}
	}

	namespace initialiser
	{
		void load_crypto_functions(temple::lua::state thread)
		{
			auto crypt_str = xorstr("crypt");
			auto subbase64_str = xorstr("base64");

			functions::lua::createtable(thread, 0, 0);

			functions::lua::createtable(thread, 0, 0);
			functions::lua::setfield(thread, -2, subbase64_str);

			functions::lua::setfield(thread, -10002, crypt_str);

			set_table_function(thread, crypt::encrypt, xorstr("encrypt"), crypt_str);
			set_table_function(thread, crypt::decrypt, xorstr("decrypt"), crypt_str);
			set_table_function(thread, crypt::random, xorstr("random"), crypt_str);
			set_subtable_function(thread, crypt::base64::encode, xorstr("encode"), crypt_str, subbase64_str);
			set_subtable_function(thread, crypt::base64::decode, xorstr("decode"), crypt_str, subbase64_str);
		}
	}
}