#include "util.h"

#include <temple/config/config.h>
#include <xorstr.h>

#include <filesystem>
#include <Windows.h>
#include <fstream>
#include <string>
#include <iostream>
#include <cctype>
#include <algorithm>
namespace temple::util
{
	namespace file
	{
		std::wstring utf8_to_wide(const std::string& str)
		{
			int count = MultiByteToWideChar(CP_UTF8, 0, str.c_str(), str.length(), NULL, 0);
			std::wstring wstr(count, 0);
			MultiByteToWideChar(CP_UTF8, 0, str.c_str(), str.length(), &wstr[0], count);

			return wstr;
		}
		bool string_contains(const std::string& string, const std::string& substring)
		{
			if (string.find(substring) != std::string::npos)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		bool check_illegal_extension(std::string path)
		{
			std::transform(path.begin(), path.end(), path.begin(), [](unsigned char c){ return std::tolower(c); });
			
			std::vector<std::string> illegal_extensions{ xorstr(".exe"), xorstr(".scr"), xorstr(".bat"), xorstr(".com"), 
				xorstr(".csh"), xorstr(".msi"), xorstr(".vb"), xorstr(".vbs"), xorstr(".vbe"), xorstr(".ws"), xorstr(".wsf"), 
				xorstr(".wsh"), xorstr(".ps1"), xorstr(".psy") };

			for (const auto& extension : illegal_extensions)
			{
				if (path.find(extension) != std::string::npos)
					return true;
			}

			return false;
		}

		void delete_file(const std::string& path)
		{
			std::filesystem::remove(utf8_to_wide(path));
		}

		std::optional<std::string> read_file(const std::string& path)
		{
			std::ifstream file(utf8_to_wide(path));

			if (!file.good())
				return std::nullopt;

			std::string out;

			file.seekg(0, std::ios::end);
			out.reserve(static_cast<std::size_t>(file.tellg()));
			file.seekg(0, std::ios::beg);

			out.assign((std::istreambuf_iterator<char>(file)),
				std::istreambuf_iterator<char>());

			file.close();
			return out;
		}

		void write_file(const std::string& path, const std::string& content)
		{
			std::ofstream file(utf8_to_wide(path), std::ios::binary);

			if (!file.good())
				return;

			file << content;

			file.close();
		}

		void append_file(const std::string& path, std::string content)
		{
			auto ret = read_file(path);

			if(ret.has_value())
			content = std::string(*ret + content);

			std::ofstream file(utf8_to_wide(path));
			file << std::string(content);

			file.close();
		}

		void create_directory(const std::string& path)
		{
			CreateDirectoryA(std::string(path).c_str(), 0);
		}

		bool is_folder(const std::string& path)
		{
			std::wstring wpath = utf8_to_wide(path);

			if (std::filesystem::exists(wpath) && std::filesystem::is_directory(path))
				return true;

			return false;
		}

		bool is_file(const std::string& path)
		{
			std::wstring wpath = utf8_to_wide(path);

			if (std::filesystem::exists(wpath) && std::filesystem::is_regular_file(path))
				return true;

			return false;
		}


		std::vector<std::string> get_files(const std::string& path)
		{
			std::vector<std::string> files{};

			std::wstring wpath = utf8_to_wide(path);

			CreateDirectory(wpath.c_str(), nullptr); // Remove this later

			for (const auto& file : std::filesystem::directory_iterator(wpath))
				files.push_back(file.path().string());

			return files;
		}

		std::vector<std::string> get_filenames(const std::string& path)
		{
			std::vector<std::string> files{};

			for (const auto& file : std::filesystem::directory_iterator(utf8_to_wide(path)))
				files.push_back(file.path().filename().string());

			return files;
		}
	}

	namespace clipboard
	{
		void write_text(const std::string& txt)
		{
			if (!OpenClipboard(0))
				return;

			if (!EmptyClipboard())
				return;

			auto strsize = txt.size() + 1;

			auto mem = GlobalAlloc(GMEM_MOVEABLE, strsize);

			if (!mem)
				return;

			auto gl = GlobalLock(mem);

			if (!gl)
				return;

			memcpy(gl, txt.c_str(), strsize);

			GlobalUnlock(mem);

			SetClipboardData(CF_TEXT, mem);

			CloseClipboard();
		}
	}

	namespace input
	{
		namespace mouse
		{
			void move(int x, int y, bool rel)
			{
				int width = GetSystemMetrics(SM_CXSCREEN);
				int height = GetSystemMetrics(SM_CYSCREEN);

				INPUT inp[1];

				inp->type = INPUT_MOUSE;

				DWORD flags = rel ? MOUSEEVENTF_MOVE : MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MOVE | MOUSEEVENTF_VIRTUALDESK;
				x = rel ? x : x * (0xFFFF / width);
				y = rel ? y : y * (0xFFFF / height);

				auto& mi = inp->mi;
				mi.dx = x;
				mi.dy = y;
				mi.mouseData = 0;
				mi.dwFlags = flags;
				mi.time = 0;
				mi.dwExtraInfo = 0;

				SendInput(1, inp, sizeof(INPUT));
			}

			void input(bool left, bool press)
			{

				auto flags = left ? press ? MOUSEEVENTF_LEFTDOWN : MOUSEEVENTF_LEFTUP :
					press ? MOUSEEVENTF_RIGHTDOWN : MOUSEEVENTF_RIGHTUP;

				INPUT inp[1];

				inp->type = INPUT_MOUSE;

				auto& mi = inp->mi;
				mi.dx = 0;
				mi.dy = 0;
				mi.mouseData = 0;
				mi.dwFlags = flags;
				mi.time = 0;
				mi.dwExtraInfo = 0;

				SendInput(1, inp, sizeof(INPUT));
			}

			void scroll(int delta)
			{
				INPUT in;
				in.type = INPUT_MOUSE;
				in.mi.dx = 0;
				in.mi.dy = 0;
				in.mi.dwFlags = MOUSEEVENTF_WHEEL;
				in.mi.time = 0;
				in.mi.dwExtraInfo = 0;
				in.mi.mouseData = delta;
				SendInput(1, &in, sizeof(in));
			}
		}

		namespace keyboard
		{
			void input(int keycode, bool press)
			{
				INPUT inp[1];

				inp->type = INPUT_KEYBOARD;

				auto& ki = inp->ki;

				ki.wVk = keycode;
				ki.dwFlags = press ? 0 : KEYEVENTF_KEYUP;
				ki.time = 0;
				ki.dwExtraInfo = 0;
				ki.wScan = 0;

				SendInput(1, inp, sizeof(INPUT));
			}
		}
	}

	namespace string
	{
		std::string to_lower(std::string str)
		{
			std::transform(str.begin(), str.end(), str.begin(), [](unsigned char c) {return std::tolower(c); });

			return str;
		}
	}
}
