#include "overlay.h"

#include <indicium/IndiciumDirect3D11.h>
#include <indicium/IndiciumDirect3D10.h>
#include <indicium/IndiciumDirect3D9.h>
#include <indicium/IndiciumCore.h>

#include <imgui/imgui.h>
#include <imgui/imgui_impl_dx11.h>
#include <imgui/imgui_impl_win32.h>
#include <xorstr.h>

#include <d3dcompiler.h>

#include <unordered_map>
#include <iostream>
#include <string>
#include <mutex>

IMGUI_IMPL_API LRESULT ImGui_ImplWin32_WndProcHandler(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

namespace temple::overlay
{
	struct temple_font
	{
		ImFont* font;
		std::string font_name;
		std::string font_path;

		temple_font(std::string name, std::string path) : font_name(name), font_path(path) {}
		temple_font() {}
	};

	std::vector<temple_font> font_load_queue;
	std::unordered_map<std::string, temple_font> fonts;

	static ID3D11RenderTargetView* g_d3d11_mainrendertargetview = nullptr;
	PINDICIUM_ENGINE engine = nullptr;

	EVT_INDICIUM_D3D11_PRE_PRESENT evt_indiciumd3d11_present;
	EVT_INDICIUM_D3D11_PRE_RESIZE_BUFFERS evt_indiciumd3d11_preresizebuffers;
	EVT_INDICIUM_D3D11_POST_RESIZE_BUFFERS evt_indiciumd3d11_postresizebuffers;

	void evt_indicium_gamehooked(PINDICIUM_ENGINE, const INDICIUM_D3D_VERSION);

	typedef LRESULT(WINAPI* t_windowproc)(_In_ HWND, _In_ UINT, _In_ WPARAM, _In_ LPARAM);


	t_windowproc original_defwindowproc = nullptr;
	t_windowproc original_windowproc = nullptr;

	bool initialise_overlay()
	{
		INDICIUM_ENGINE_CONFIG config;
		INDICIUM_ENGINE_CONFIG_INIT(&config);

		config.Direct3D.HookDirect3D9 = true;
		config.Direct3D.HookDirect3D10 = true;
		config.Direct3D.HookDirect3D11 = true;
		config.EvtIndiciumGameHooked = evt_indicium_gamehooked;

		auto result = IndiciumEngineCreate(GetModuleHandle(nullptr), &config, nullptr);

		if (result == INDICIUM_ERROR::INDICIUM_ERROR_NONE)
			return true;
		else
			return false;
	}

	void evt_indicium_gamehooked(PINDICIUM_ENGINE engine_handle, const INDICIUM_D3D_VERSION game_version)
	{

		IMGUI_CHECKVERSION();
		ImGui::CreateContext();
		auto& io = ImGui::GetIO();
		//io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;  // Enable Keyboard Controls
		//io.ConfigFlags |= ImGuiConfigFlags_NavEnableGamepad;   // Enable Gamepad Controls

		// Setup Dear ImGui style
		ImGui::StyleColorsDark();
		//ImGui::StyleColorsClassic();

		INDICIUM_D3D11_EVENT_CALLBACKS d3d11;
		INDICIUM_D3D11_EVENT_CALLBACKS_INIT(&d3d11);
		d3d11.EvtIndiciumD3D11PrePresent = evt_indiciumd3d11_present;
		d3d11.EvtIndiciumD3D11PreResizeBuffers = evt_indiciumd3d11_preresizebuffers;
		d3d11.EvtIndiciumD3D11PostResizeBuffers = evt_indiciumd3d11_postresizebuffers;



		switch (game_version)
		{
		case IndiciumDirect3DVersion11:
			IndiciumEngineSetD3D11EventCallbacks(engine_handle, &d3d11);
			break;
		}
	}

	void load_font(std::string name, std::string path)
	{
		const auto& it = fonts.find(name);

		//std::cout << name << " " << path << "\n";

		if (it == fonts.end())
		{
			path = xorstr("C:\\Windows\\Fonts\\") + path;

			//std::cout << path << "\n";

			font_load_queue.push_back(temple_font(name, path));
		}
	}

	void text_set_bounds(elements::text* text)
	{
		////std::cout << text->Font << "\n";
		const auto& it = fonts.find(text->font);

		ImFont* font = nullptr;

		if (it != fonts.end())
		{
			////std::cout << it->first << "\n";
			font = it->second.font;
		}

		ImVec2 size = font->CalcTextSizeA(text->size, FLT_MAX, 0.0f, text->content.c_str());

		text->text_bounds = elements::vector2(size.x, size.y);
	}
	
	void render_scene()
	{
		auto draw_list = ImGui::GetOverlayDrawList();

		for (elements::base* element : elements::elements)
		{
			if (!element->visible)
				continue;

			switch (element->get_type())
			{
			case elements::element_type::line:
			{
				auto line = static_cast<elements::line*>(element);

				draw_list->AddLine(ImVec2(line->p1.x, line->p1.y), ImVec2(line->p2.x, line->p2.y), line->color.to_hex(), line->thickness);
				break;
			}
			case elements::element_type::circle:
			{
				auto circle = static_cast<elements::circle*>(element);

				if (circle->filled)
					draw_list->AddCircleFilled(ImVec2(circle->position.x, circle->position.y), circle->radius, circle->color.to_hex(), circle->edges);
				else
					draw_list->AddCircle(ImVec2(circle->position.x, circle->position.y), circle->radius, circle->color.to_hex(), circle->edges, circle->thickness);

				break;
			}
			case elements::element_type::square:
			{
				auto square = static_cast<elements::square*>(element);

				if (square->filled)
					draw_list->AddRectFilled(ImVec2(square->position.x, square->position.y), ImVec2(
						square->position.x + square->size.x, square->position.y + square->size.y), square->color.to_hex());
				else
					draw_list->AddRect(ImVec2(square->position.x, square->position.y), ImVec2(
						square->position.x + square->size.x, square->position.y + square->size.y), square->color.to_hex(), 0, 15, square->thickness);

				break;
			}
			case elements::element_type::triangle:
			{
				auto triangle = static_cast<elements::triangle*>(element);

				if (triangle->filled)
					draw_list->AddTriangleFilled(ImVec2(triangle->p1.x, triangle->p1.y), ImVec2(triangle->p2.x, triangle->p2.y),
						ImVec2(triangle->p3.x, triangle->p3.y), triangle->color.to_hex());
				else
					draw_list->AddTriangle(ImVec2(triangle->p1.x, triangle->p1.y), ImVec2(triangle->p2.x, triangle->p2.y),
						ImVec2(triangle->p3.x, triangle->p3.y), triangle->color.to_hex(), triangle->thickness);

				break;
			}
			case elements::element_type::quad:
			{
				auto quad = static_cast<elements::quad*>(element);

				if (quad->filled)
					draw_list->AddQuadFilled(ImVec2(quad->p1.x, quad->p1.y), ImVec2(quad->p2.x, quad->p2.y),
						ImVec2(quad->p3.x, quad->p3.y), ImVec2(quad->p4.x, quad->p4.y), quad->color.to_hex());
				else
					draw_list->AddQuad(ImVec2(quad->p1.x, quad->p1.y), ImVec2(quad->p2.x, quad->p2.y),
						ImVec2(quad->p3.x, quad->p3.y), ImVec2(quad->p4.x, quad->p4.y), quad->color.to_hex(), quad->thickness);

				break;
			}
			case elements::element_type::text:
			{
				auto text = static_cast<elements::text*>(element);
				
				const auto& it = fonts.find(text->font);
				bool pushed = false;

				if (it != fonts.end())
				{
					////std::cout << it->first << "\n";
					auto& font = it->second;

					ImGui::PushFont(font.font);
					pushed = true;
				}

				float outlineextra = 0;

				ImVec2 size = ImGui::GetFont()->CalcTextSizeA(text->size, FLT_MAX, 0.0f, text->content.c_str());

				text->text_bounds = elements::vector2(size.x, size.y);

				auto position = text->position;

				if (text->center)
				{
					position.x -= text->text_bounds.x / 2;
				}

				/*
				if (text->outline)
				{
					draw_list->AddText(NULL, text->size + 0.5, ImVec2(text->position.x + extrapos, text->position.y), 0xFF000000,
						text->content.c_str(), text->content.c_str() + (text->content.length() * sizeof(char)));
					outlineextra = 0.5;
				}*/

				if (text->outline)
				{
					auto outline_color = elements::color3(text->color.a, text->outline_color.r, text->outline_color.g, text->outline_color.b);
					auto hex_color = outline_color.to_hex();

					draw_list->AddText(NULL, text->size, ImVec2(position.x + 1, position.y + 1),
						hex_color, text->content.c_str());

					draw_list->AddText(NULL, text->size, ImVec2(position.x - 1, position.y - 1),
						hex_color, text->content.c_str());

					draw_list->AddText(NULL, text->size, ImVec2(position.x + 1, position.y - 1),
						hex_color, text->content.c_str());

					draw_list->AddText(NULL, text->size, ImVec2(position.x - 1, position.y + 1),
						hex_color, text->content.c_str());
				}

				draw_list->AddText(NULL, text->size, ImVec2(position.x, position.y),
					text->color.to_hex(), text->content.c_str());

				if (pushed)
					ImGui::PopFont();



				break;
			}
			}
		}

		ImGui::Render();
	}

	void evt_indiciumd3d11_present(IDXGISwapChain* swap_chain, UINT sync_interval, UINT flags, PINDICIUM_EVT_PRE_EXTENSION extension)
	{
		static auto initialized = false;
		static bool show_overlay = true;
		static std::once_flag init;

		static ID3D11DeviceContext* context;

		//
		// This section is only called once to initialize ImGui
		// 
		std::call_once(init, [&](IDXGISwapChain* chain)
			{

				ID3D11Device* device;
				if (FAILED(D3D11_DEVICE_IMMEDIATE_CONTEXT_FROM_SWAPCHAIN(chain, &device, &context)))
				{
					return;
				}

				ID3D11Texture2D* back_buffer;
				chain->GetBuffer(0, __uuidof(ID3D11Texture2D), (LPVOID*)&back_buffer);
				device->CreateRenderTargetView(back_buffer, NULL, &g_d3d11_mainrendertargetview);
				back_buffer->Release();

				DXGI_SWAP_CHAIN_DESC sd;
				chain->GetDesc(&sd);

				ImGui_ImplWin32_Init(sd.OutputWindow);
				ImGui_ImplDX11_Init(device, context);

				//ID3D11PixelShader* PS;
				//ID3DBlob* PS_Buffer;

				//D3DReadFileToBlob(L"C:\\Users\\dicue\\Desktop\\vsSSR.cso", &PS_Buffer);
				//device->CreatePixelShader(PS_Buffer->GetBufferPointer(), PS_Buffer->GetBufferSize(), NULL, &PS);
				//context->PSSetShader(PS, 0, 0);

				//HookWindowProc(sd.OutputWindow);

				initialized = true;

			}, swap_chain);

		if (!initialized)
			return;

		//TOGGLE_STATE(VK_F12, show_overlay);
		//if (!show_overlay)
		//	return;

		for (size_t i = 0; i < font_load_queue.size(); i++)
		{
			temple_font font = font_load_queue[i];

			ImGuiIO& io = ImGui::GetIO();

			font.font = io.Fonts->AddFontFromFileTTF(font.font_path.c_str(), 32);
			io.Fonts->Build();
			ImGui_ImplDX11_InvalidateDeviceObjects();
			//io.Fonts->GetTexDataAsRGBA32()

			fonts[font.font_name] = font;
		}

		font_load_queue.clear();

		// Start the Dear ImGui frame
		ImGui_ImplDX11_NewFrame();
		ImGui_ImplWin32_NewFrame();
		ImGui::NewFrame();

		context->OMSetRenderTargets(1, &g_d3d11_mainrendertargetview, NULL);

		render_scene();


		ImGui_ImplDX11_RenderDrawData(ImGui::GetDrawData());
	}

	void evt_indiciumd3d11_preresizebuffers(IDXGISwapChain* swap_chain, UINT buffer_count, UINT width, UINT height,
		DXGI_FORMAT new_format, UINT swap_chain_flags, PINDICIUM_EVT_PRE_EXTENSION extension)
	{
		if (g_d3d11_mainrendertargetview)
		{
			g_d3d11_mainrendertargetview->Release();
			g_d3d11_mainrendertargetview = nullptr;
		}
	}

	void evt_indiciumd3d11_postresizebuffers(IDXGISwapChain* swap_chain, UINT buffer_count, UINT width, UINT height,
		DXGI_FORMAT new_format, UINT swap_chain_flags, PINDICIUM_EVT_POST_EXTENSION extension)
	{
		ID3D11Texture2D* back_buffer;
		ID3D11DeviceContext* context;
		ID3D11Device* device;
		D3D11_DEVICE_IMMEDIATE_CONTEXT_FROM_SWAPCHAIN(swap_chain, &device, &context);

		swap_chain->GetBuffer(0, IID_PPV_ARGS(&back_buffer));
		device->CreateRenderTargetView(back_buffer, NULL, &g_d3d11_mainrendertargetview);
		back_buffer->Release();
	}

	LRESULT WINAPI detour_defwindowproc(_In_ HWND hwnd, _In_ UINT msg, _In_ WPARAM wparam,_In_ LPARAM lparam)
	{
		ImGui_ImplWin32_WndProcHandler(hwnd, msg, wparam, lparam);

		return original_defwindowproc(hwnd, msg, wparam, lparam);
	}

	LRESULT WINAPI detourwindowproc(_In_ HWND hwnd, _In_ UINT msg, _In_ WPARAM wparam,_In_ LPARAM lparam)
	{
		ImGui_ImplWin32_WndProcHandler(hwnd, msg, wparam, lparam);

		return original_windowproc(hwnd, msg, wparam, lparam);
	}
}
