#include "crypto.h"

namespace temple::crypto
{
	std::string bin2base64(const std::string& bin)
	{
		std::size_t base64size = sodium_base64_ENCODED_LEN(bin.length(), sodium_base64_VARIANT_ORIGINAL);

		char* base64res = new char[base64size];
		sodium_bin2base64(base64res, base64size, reinterpret_cast<unsigned char*>(const_cast<char*>(bin.c_str())), bin.length(), sodium_base64_VARIANT_ORIGINAL);

		std::string result(base64res);

		delete[] base64res;

		return result;
	}

	std::string base642bin(const std::string& base64)
	{
		std::size_t base64size = base64.length();

		size_t binsize = 0;
		const char* b64end = 0;
		unsigned char* binres = new unsigned char[base64size];
		sodium_base642bin(binres, base64size, base64.c_str(), base64.length(), "", &binsize, &b64end, sodium_base64_VARIANT_ORIGINAL);

		std::string result(reinterpret_cast<char*>(binres), binsize);

		return result;
	}

	std::string from_hex(const std::string& input)
	{
		std::string hex_table = "0123456789abcdef";
		std::string buffer;

		buffer.reserve(input.size() / 2);

		for (std::size_t i = 0; i < input.size(); i += 2)
			buffer += static_cast<char>((hex_table.find(input[i]) << 4) | hex_table.find(input[i + 1]));

		return buffer;
	}
	std::string to_hex(const std::string& input)
	{
		std::string hex_table = "0123456789abcdef";
		std::string buffer;

		buffer.reserve(input.size() * 2);

		for (const auto& c : input)
		{
			buffer += hex_table[static_cast<unsigned char>(c) >> 4];
			buffer += hex_table[static_cast<unsigned char>(c) & 15];
		}

		return buffer;
	}

	std::string sha256(const std::string& input)
	{
		crypto_string buffer{crypto_hash_sha256_BYTES};
		crypto_hash_sha256(buffer.buffer, reinterpret_cast<const unsigned char*>(input.c_str()), input.size());

		return buffer;
	}

	std::string sha512(const std::string& input)
	{
		crypto_string buffer{ crypto_hash_sha512_BYTES };
		crypto_hash_sha512(buffer.buffer, reinterpret_cast<const unsigned char*>(input.c_str()), input.size());

		return buffer;
	}

	std::string random_bytes(size_t len)
	{
		crypto_string buffer(len);
		randombytes_buf(buffer.buffer, buffer.length);

		return buffer;
	}

	std::string signature::sign(const std::string& message) const
	{
		crypto_string sig(crypto_sign_BYTES);
		crypto_sign_detached(sig.buffer, nullptr, reinterpret_cast<const unsigned char*>(message.c_str()),
			message.size(), secret_key);

		return sig;
	}

	bool signature::verify(const std::string& message, const std::string& sig, const signature& pkey)
	{
		return crypto_sign_verify_detached(reinterpret_cast<const unsigned char*>(sig.c_str()), 
			reinterpret_cast<const unsigned char*>(message.c_str()), message.size(), pkey.public_key) == 0;
	}

	std::string pk_encryption::encrypt(const std::string& input) const
	{
		auto rnd_buffer = random_bytes(16) + input;
		crypto_string message(crypto_box_SEALBYTES + rnd_buffer.size());

		crypto_box_seal(message.buffer, reinterpret_cast<const unsigned char*>(rnd_buffer.c_str()), 
			rnd_buffer.size(), public_key);

		return message;
	}

	std::optional<std::string> pk_encryption::decrypt(const std::string& input) const
	{
		crypto_string decrypted(input.size() - crypto_box_SEALBYTES);

		auto success = crypto_box_seal_open(decrypted.buffer, reinterpret_cast<const unsigned char*>(input.c_str()), 
			input.size(), public_key, secret_key) == 0;

		if (success)
			return static_cast<std::string>(decrypted).substr(16);
		else
			return std::nullopt;
	}

	namespace lua 
	{
		/*
		[[nodiscard]] unsigned char* ref(std::string& p)
		{
			return reinterpret_cast<unsigned char*>(p->c_str());
		}

		[[nodiscard]] std::string nonce(std::size_t nonce_size) 
		{
			std::string nbuf (nonce_size, ' ');
			randombytes_buf(ref(nbuf), nonce_size);

			return nbuf;
		};

		template<std::size_t key_size>
		[[nodiscard]] std::string resize_key(const std::string& key)
		{
			if (key.length() >= key_size)
			{
				return key.substr(0, key_size);
			}

			return key + std::string(key_size - key.size(), '\x00');
		}
		namespace aes256gcm {

			static constexpr auto key_length = crypto_aead_aes256_gcm_KEYBYTES;
			static constexpr auto nonce_length = crypto_aead_aes256gcm_NPUBBYTES;

			[[nodiscard]] std::string encrypt(const std::string& message, const std::string& unfiltered_key)
			{
				if (crypto_aead_aes256gcm_is_available() == 0)
				{
					throw std::runtime_error("aes256gcm algorithm unavailable on this CPU");
				}

				auto key = resize_key<key_length>(unfiltered_key);
				auto n = nonce(nonce_length);

				unsigned long long data_len = message.length() + crypto_aead_aes256gcm_ABYTES;

				std::string data;
				data.resize(data_len);

				crypto_aead_aes256gcm_encrypt(ref(data), &data_len, ref(message), message.length(), nullptr, 0, ref(n), ref(key));

				return data;
			}

			[[nodiscard]] std::string decrypt(const std::string& data, const std::string& unfiltered_key)
			{
				if (crypto_aead_aes256gcm_is_available() == 0)
				{
					throw std::runtime_error("aes256gcm algorithm unavailable on this CPU");
				}

				if (data.length() < crypto_aead_aes256gcm_ABYTES)
				{
					throw std::runtime_error("Invalid data length passed to aes256gcm");
				}

				auto key = resize_key<key_length>(unfiltered_key);

				unsigned long long message_len = data.length() - crypto_aead_aes256gcm_ABYTES; 

				std::string message;
				message.resize(message_len);

				std::string nonce;
				nonce.resize(nonce_length);

				crypto_aead_aes256gcm_decrypt(ref(message), &message_len, nullptr, ref(data), data.length(), nullptr, 0, ref(nonce), ref(key));

				return message;
			}

		}*/
	}
}