#pragma once
#include <libsodium/sodium.h>

#include <optional>
#include <cstddef>
#include <string>

namespace temple::crypto
{
	struct crypto_string
	{
		unsigned char* buffer;
		std::size_t length;

		operator std::string()
		{
			return std::string(reinterpret_cast<char*>(buffer), length);
		}

		~crypto_string()
		{
			delete[] buffer;
		}

		crypto_string(std::size_t length) : length(length), buffer(new unsigned char[length]) {}
	};

	template<std::size_t public_size, std::size_t secret_size>
	struct key_pair
	{
		unsigned char secret_key[secret_size];
		unsigned char public_key[public_size];

		key_pair() {}

		key_pair(const std::string& pbl_key, const std::string& scr_key = "")
		{
			std::memcpy(public_key, pbl_key.c_str(), public_size);
			std::memcpy(secret_key, scr_key.c_str(), secret_size);
		}

		auto get_public()
		{
			return std::string(reinterpret_cast<char*>(public_key), public_size);
		}
		auto get_secret()
		{
			return std::string(reinterpret_cast<char*>(secret_key), secret_size);
		}
	};

	struct signature: public key_pair<crypto_sign_PUBLICKEYBYTES, crypto_sign_SECRETKEYBYTES>
	{
		static bool verify(const std::string&, const std::string&, const signature&);
		std::string sign(const std::string&) const;

		signature()
		{
			crypto_sign_keypair(public_key, secret_key);
		}

		signature(const std::string& public_key, const std::string& secret_key = "") :
			key_pair(public_key, secret_key) {};
	};

	struct pk_encryption : public key_pair<crypto_box_PUBLICKEYBYTES, crypto_box_SECRETKEYBYTES>
	{
		std::string encrypt(const std::string&) const;
		std::optional<std::string> decrypt(const std::string&) const;

		pk_encryption()
		{
			crypto_box_keypair(public_key, secret_key);
		}

		pk_encryption(const std::string& public_key, const std::string& secret_key = "") : key_pair(public_key, secret_key) {}
	};

	std::string bin2base64(const std::string&);
	std::string base642bin(const std::string&);
	std::string sha256(const std::string&);
	std::string sha512(const std::string&);
	std::string from_hex(const std::string&);
	std::string to_hex(const std::string&);

	namespace lua 
	{
		[[nodiscard]] unsigned char* ref(std::string& p);

		[[nodiscard]] std::string nonce(std::size_t nonce_size);

		namespace aes256gcm {

			[[nodiscard]] std::string encrypt(const std::string& message, const std::string& unfiltered_key);

			[[nodiscard]] std::string decrypt(const std::string& data, const std::string& unfiltered_key);

		}
	}
}